import os
import tempfile

import matplotlib.pyplot as plt
import numpy as np
import pyautogui
from PIL import Image


class Display:
    def screenshot(self, show=True, quadrant=None):
        temp_file = tempfile.NamedTemporaryFile(suffix=".png", delete=False)

        if quadrant == None:
            screenshot = pyautogui.screenshot()
        else:
            try:
                screen_width, screen_height = pyautogui.size()
            except:
                raise EnvironmentError("Unable to determine screen size.")

            quadrant_width = screen_width // 2
            quadrant_height = screen_height // 2

            quadrant_coordinates = {
                1: (0, 0),
                2: (quadrant_width, 0),
                3: (0, quadrant_height),
                4: (quadrant_width, quadrant_height),
            }

            if quadrant in quadrant_coordinates:
                x, y = quadrant_coordinates[quadrant]
                screenshot = pyautogui.screenshot(
                    region=(x, y, quadrant_width, quadrant_height)
                )
            else:
                raise ValueError("Invalid quadrant. Choose between 1 and 4.")

        screenshot.save(temp_file.name)

        # Open the image file with PIL
        img = Image.open(temp_file.name)

        # Delete the temporary file
        os.remove(temp_file.name)

        if show:
            # Show the image using matplotlib
            plt.imshow(np.array(img))
            plt.show()

        return img



from .terminal.terminal import Terminal

try:
    from .clipboard.clipboard import Clipboard
    from .display.display import Display
    from .keyboard.keyboard import Keyboard
    from .mouse.mouse import Mouse
except:
    pass


class Computer:
    def __init__(self):
        self.terminal = Terminal()
        try:
            self.mouse = Mouse(
                self
            )  # Mouse will use the computer's display, so we give it a reference to ourselves
            self.keyboard = Keyboard()
            self.display = Display()
            self.clipboard = Clipboard()
        except:
            pass

    def run(self, *args, **kwargs):
        """
        Shortcut for computer.terminal.run
        """
        return self.terminal.run(*args, **kwargs)

    def stop(self):
        """
        Shortcut for computer.terminal.stop
        """
        return self.terminal.stop()

    def terminate(self):
        """
        Shortcut for computer.terminal.terminate
        """
        return self.terminal.terminate()

    def screenshot(self, *args, **kwargs):
        """
        Shortcut for computer.display.screenshot
        """
        return self.display.screenshot(*args, **kwargs)



import traceback

from .rag.get_relevant_procedures_string import get_relevant_procedures_string
from .utils.get_user_info_string import get_user_info_string


def generate_system_message(interpreter):
    """
    Dynamically generate a system message.

    Takes an interpreter instance,
    returns a string.

    This is easy to replace!
    Just swap out `interpreter.generate_system_message` with another function.
    """

    #### Start with the static system message

    system_message = interpreter.system_message

    #### Add dynamic components, like the user's OS, username, relevant procedures, etc

    system_message += "\n" + get_user_info_string()

    # DISABLED
    # because wait, they'll have terminal open, no text will be selected. if we find a way to call `--os` mode from anywhere, this will be cool though.
    # if interpreter.os:
    #     # Add the user's selection to to the system message in OS mode
    #     try:
    #         selected_text = interpreter.computer.clipboard.get_selected_text()
    #         if len(selected_text) > 20:
    #             system_message += "\nThis is a preview of the user's selected text: " + selected_text[:20] + "..." + selected_text[-20:]
    #     except:
    #         pass

    if not interpreter.local and not interpreter.disable_procedures:
        try:
            system_message += "\n" + get_relevant_procedures_string(interpreter)
        except:
            if interpreter.debug_mode:
                print(traceback.format_exc())
            # It's okay if they can't. This just fixes some common mistakes it makes.

    for language in interpreter.computer.terminal.languages:
        if hasattr(language, "system_message"):
            system_message += "\n\n" + language.system_message

    return system_message.strip()



import argparse
import os
import platform
import re
import subprocess
import sys

import pkg_resources

from .conversation_navigator import conversation_navigator
from .check_for_update import check_for_update
from .display_markdown_message import display_markdown_message
from .get_config import get_config_path
from .validate_llm_settings import validate_llm_settings

arguments = [
    {
        "name": "system_message",
        "nickname": "s",
        "help_text": "prompt / custom instructions for the language model",
        "type": str,
    },
    {
        "name": "local",
        "nickname": "l",
        "help_text": "experimentally run the language model locally (via LM Studio)",
        "type": bool,
    },
    {
        "name": "auto_run",
        "nickname": "y",
        "help_text": "automatically run generated code",
        "type": bool,
    },
    {
        "name": "debug_mode",
        "nickname": "d",
        "help_text": "run in debug mode",
        "type": bool,
    },
    {
        "name": "disable_procedures",
        "nickname": "dp",
        "help_text": "disables procedures (RAG of some common OI use-cases). disable to shrink system message. auto-disabled for non-OpenAI models",
        "type": bool,
    },
    {
        "name": "model",
        "nickname": "m",
        "help_text": "language model to use",
        "type": str,
    },
    {
        "name": "temperature",
        "nickname": "t",
        "help_text": "optional temperature setting for the language model",
        "type": float,
    },
    {
        "name": "context_window",
        "nickname": "c",
        "help_text": "optional context window size for the language model",
        "type": int,
    },
    {
        "name": "max_tokens",
        "nickname": "x",
        "help_text": "optional maximum number of tokens for the language model",
        "type": int,
    },
    {
        "name": "max_output",
        "nickname": "xo",
        "help_text": "optional maximum number of characters for code outputs",
        "type": int,
    },
    {
        "name": "max_budget",
        "nickname": "b",
        "help_text": "optionally set the max budget (in USD) for your llm calls",
        "type": float,
    },
    {
        "name": "api_base",
        "nickname": "ab",
        "help_text": "optionally set the API base URL for your llm calls (this will override environment variables)",
        "type": str,
    },
    {
        "name": "api_key",
        "nickname": "ak",
        "help_text": "optionally set the API key for your llm calls (this will override environment variables)",
        "type": str,
    },
    {
        "name": "safe_mode",
        "nickname": "safe",
        "help_text": "optionally enable safety mechanisms like code scanning; valid options are off, ask, and auto",
        "type": str,
        "choices": ["off", "ask", "auto"],
        "default": "off",
    },
    {
        "name": "config_file",
        "nickname": "cf",
        "help_text": "optionally set a custom config file to use",
        "type": str,
    },
    {
        "name": "vision",
        "nickname": "v",
        "help_text": "experimentally use vision for supported languages (HTML)",
        "type": bool,
    },
    {
        "name": "os",
        "nickname": "o",
        "help_text": "experimentally let RobotGPT Interpreter control your mouse and keyboard",
        "type": bool,
    },
]


def start_terminal_interface(interpreter):
    """
    Meant to be used from the command line. Parses arguments, starts OI's terminal interface.
    """

    parser = argparse.ArgumentParser(description="RobotGPT Interpreter")

    # Add arguments
    for arg in arguments:
        if arg["type"] == bool:
            parser.add_argument(
                f'-{arg["nickname"]}',
                f'--{arg["name"]}',
                dest=arg["name"],
                help=arg["help_text"],
                action="store_true",
                default=None,
            )
        else:
            choices = arg["choices"] if "choices" in arg else None
            default = arg["default"] if "default" in arg else None

            parser.add_argument(
                f'-{arg["nickname"]}',
                f'--{arg["name"]}',
                dest=arg["name"],
                help=arg["help_text"],
                type=arg["type"],
                choices=choices,
                default=default,
            )

    # Add special arguments
    parser.add_argument(
        "--config",
        dest="config",
        action="store_true",
        help="open config.yaml file in text editor",
    )
    parser.add_argument(
        "--conversations",
        dest="conversations",
        action="store_true",
        help="list conversations to resume",
    )
    parser.add_argument(
        "-f",
        "--fast",
        dest="fast",
        action="store_true",
        help="run `interpreter --model gpt-3.5-turbo`",
    )
    parser.add_argument(
        "--version",
        dest="version",
        action="store_true",
        help="get RobotGPT Interpreter's version number",
    )

    args = parser.parse_args()

    # This should be pushed into an open_config.py util
    # If --config is used, open the config.yaml file in the RobotGPT Interpreter folder of the user's config dir
    if args.config:
        if args.config_file:
            config_file = get_config_path(args.config_file)
        else:
            config_file = get_config_path()

        print(f"Opening `{config_file}`...")

        # Use the default system editor to open the file
        if platform.system() == "Windows":
            os.startfile(
                config_file
            )  # This will open the file with the default application, e.g., Notepad
        else:
            try:
                # Try using xdg-open on non-Windows platforms
                subprocess.call(["xdg-open", config_file])
            except FileNotFoundError:
                # Fallback to using 'open' on macOS if 'xdg-open' is not available
                subprocess.call(["open", config_file])
        return

    if args.local:
        # Default local (LM studio) attributes
        interpreter.system_message = "You are an AI."
        interpreter.model = (
            "openai/" + interpreter.model
        )  # This tells LiteLLM it's an OpenAI compatible server
        interpreter.api_base = "https://dataai.harix.iamidata.com/llm/api/ask"
        interpreter.max_tokens = 1000
        interpreter.context_window = 3000
        interpreter.api_key = "0"

        display_markdown_message(
            """
> RobotGPT Interpreter's local mode is powered by **`LM Studio`**.


You will need to run **LM Studio** in the background.

1. Download **LM Studio** from [https://lmstudio.ai/](https://lmstudio.ai/) then start it.
2. Select a language model then click **Download**.
3. Click the **<->** button on the left (below the chat button).
4. Select your model at the top, then click **Start Server**.


Once the server is running, you can begin your conversation below.

> **Warning:** This feature is highly experimental.
> Don't expect `gpt-3.5` / `gpt-4` level quality, speed, or reliability yet!

"""
        )

    # Set attributes on interpreter
    for attr_name, attr_value in vars(args).items():
        # Ignore things that aren't possible attributes on interpreter
        if attr_value is not None and hasattr(interpreter, attr_name):
            # If the user has provided a config file, load it and extend interpreter's configuration
            if attr_name == "config_file":
                user_config = get_config_path(attr_value)
                interpreter.config_file = user_config
                interpreter.extend_config(config_path=user_config)
            else:
                setattr(interpreter, attr_name, attr_value)

    # Check for update
    try:
        if not interpreter.local:
            # This should actually be pushed into the utility
            if check_for_update():
                display_markdown_message(
                    "> **A new version of RobotGPT Interpreter is available.**\n>Please run: `pip install --upgrade robotgpt-interpreter`\n\n---"
                )
    except:
        # Doesn't matter
        pass

    # if safe_mode and auto_run are enabled, safe_mode disables auto_run
    if interpreter.auto_run and (
        interpreter.safe_mode == "ask" or interpreter.safe_mode == "auto"
    ):
        setattr(interpreter, "auto_run", False)

    # If --conversations is used, run conversation_navigator
    if args.conversations:
        conversation_navigator(interpreter)
        return

    if args.version:
        version = pkg_resources.get_distribution("robotgpt-interpreter").version
        update_name = "New Computer"  # Change this with each major update
        print(f"RobotGPT Interpreter {version} {update_name}")
        return

    if args.fast:
        interpreter.model = "gpt-3.5-turbo"

    if args.vision:
        interpreter.vision = True
        interpreter.model = "gpt-4-vision-preview"
        interpreter.system_message += "\nThe user will show you an image of the code you write. You can view images directly.\n\nFor HTML: This will be run STATELESSLY. You may NEVER write '<!-- previous code here... --!>' or `<!-- header will go here -->` or anything like that. It is CRITICAL TO NEVER WRITE PLACEHOLDERS. Placeholders will BREAK it. You must write the FULL HTML CODE EVERY TIME. Therefore you cannot write HTML piecemeal—write all the HTML, CSS, and possibly Javascript **in one step, in one code block**. The user will help you review it visually.\nIf the user submits a filepath, you will also see the image. The filepath and user image will both be in the user's message.\n\nIf you use `plt.show()`, the resulting image will be sent to you. However, if you use `PIL.Image.show()`, the resulting image will NOT be sent to you."
        interpreter.function_calling_llm = False
        interpreter.context_window = 110000
        interpreter.max_tokens = 4096
        interpreter.force_task_completion = True

        display_markdown_message("> `Vision` enabled **(experimental)**\n")

    if args.os:
        interpreter.os = True
        interpreter.disable_procedures = True
        interpreter.vision = True
        interpreter.model = "gpt-4-vision-preview"
        interpreter.function_calling_llm = False
        interpreter.context_window = 110000
        interpreter.max_tokens = 4096
        interpreter.auto_run = True
        interpreter.force_task_completion = True
        # This line made it use files too much
        interpreter.system_message = interpreter.system_message.replace(
            "If you want to send data between programming languages, save the data to a txt or json.\n",
            "",
        )
        interpreter.system_message = interpreter.system_message.replace(
            "When a user refers to a filename, they're likely referring to an existing file in the directory you're currently executing code in.",
            "The user is likely referring to something on their screen.",
        )
        interpreter.system_message += (
            "\n\n"
            + """

Execute code using `computer` (already imported) to control the user's computer:

```python
computer.screenshot() # Automatically runs plt.show() to show you what's on the screen, returns a `pil_image` `in case you need it (rarely). **You almost always want to do this first! You don't know what's on the user's screen.**
computer.screenshot(quadrant=1) # Get a detailed view of the upper left quadrant (you'll rarely need this, use it to examine/retry failed attempts)

computer.keyboard.hotkey("space", "command") # Opens spotlight (very useful)
computer.keyboard.write("hello")
# .down() .up() and .press() also work (uses pyautogui)

computer.mouse.move("Text Onscreen") # This moves the mouse to the UI element with that text. Use this **frequently** — and get creative! To mouse over a video thumbnail, you could pass the *timestamp* (which is usually written on the thumbnail) into this.
computer.mouse.move(x=500, y=500) # Use this very rarely. It's only 1% as accurate as click("Text")!
computer.mouse.click() # Don't forget this! Include in the same code block

# Dragging
computer.mouse.move("So I was")
computer.mouse.down()
computer.mouse.move("and that's it!")
computer.mouse.up()
```

For rare and complex mouse actions, consider using computer vision libraries on `pil_image` to produce a list of coordinates for the mouse to move/drag to.

If the user highlighted text in an editor, then asked you to modify it, they probably want you to `keyboard.write` it over their version of the text.

Tasks are 100% computer-based. DO NOT simply write long messages to the user to complete tasks. You MUST put your text back into the program they're using to deliver your text! For example, overwriting some text they've highlighted with `keyboard.write`.

Use keyboard navigation when reasonably possible, but not if it involves pressing a button multiple times. The mouse is less reliable. Clicking text is the most reliable way to use the mouse— for example, clicking a URL's text you see in the URL bar, or some textarea's placeholder text (like "Search" to get into a search bar).

Applescript might be best for some tasks.
        
If you use `plt.show()`, the resulting image will be sent to you. However, if you use `PIL.Image.show()`, the resulting image will NOT be sent to you.

**Include `computer.screenshot()` after a 2 second delay at the end of _every_ code block to verify your progress on the task.**

Try multiple methods before saying the task is impossible. **You can do it!**

You are an expert computer navigator, brilliant and technical. **At each step, describe the user's screen with a lot of detail, including 1. the active app, 2. what text areas appear to be active, 3. what text is selected, if any, 4. what options you could take next.** Think carefully.

        """.strip()
        )

        # Download required packages
        try:
            import cv2
            import IPython
            import matplotlib
            import pyautogui
            import pytesseract
        except ImportError:
            display_markdown_message(
                "> **Missing Packages**\n\nSeveral packages are required for OS Control (`matplotlib`, `pytesseract`, `pyautogui`, `opencv-python`, `ipython`).\n\nInstall them?\n"
            )
            user_input = input("(y/n) > ")
            if user_input.lower() != "y":
                print("Exiting...")
                return
            packages = [
                "matplotlib",
                "pytesseract",
                "pyautogui",
                "opencv-python",
                "ipython",
            ]
            command = "\n".join([f"pip install {package}" for package in packages])
            for chunk in interpreter.computer.run("shell", command):
                if chunk.get("format") != "active_line":
                    print(chunk.get("content"))

        display_markdown_message(
            "> `OS Control` enabled (experimental)\n\nRobotGPT Interpreter will be able to see your screen, move your mouse, and use your keyboard."
        )
        print("")  # < - Aesthetic choice

        # Give it access to the computer via Python
        for _ in interpreter.computer.run(
            "python",
            "import interpreter\ncomputer = interpreter.computer",
        ):
            pass

        display_markdown_message(
            "**Warning:** In this mode, RobotGPT Interpreter will not require approval before performing actions. Be ready to close your terminal."
        )
        print("")  # < - Aesthetic choice

    if not interpreter.local and interpreter.model == "gpt-4-1106-preview":
        if interpreter.context_window is None:
            interpreter.context_window = 128000
        if interpreter.max_tokens is None:
            interpreter.max_tokens = 4096
        if interpreter.function_calling_llm is None:
            interpreter.function_calling_llm = True

    if not interpreter.local and interpreter.model == "gpt-3.5-turbo-1106":
        if interpreter.context_window is None:
            interpreter.context_window = 16000
        if interpreter.max_tokens is None:
            interpreter.max_tokens = 4096
        if interpreter.function_calling_llm is None:
            interpreter.function_calling_llm = True

    validate_llm_settings(interpreter)

    interpreter.chat()



import io

import cv2
import numpy as np
from PIL import Image


def find_svg_in_image(svg_code, pil_image):
    """
    Not implemented
    """
    png_image = cairosvg.svg2png(bytestring=svg_code)
    svg_image = Image.open(io.BytesIO(png_image))
    opencv_image = cv2.cvtColor(np.array(pil_image), cv2.COLOR_RGB2GRAY)

    svg_image_array = np.array(svg_image)
    if np.any(svg_image_array[:, :, 3] < 255):
        non_transparent_pixels = svg_image_array[svg_image_array[:, :, 3] > 0]
        avg_color = np.mean(non_transparent_pixels, axis=0)

        dist_black = 1 - np.linalg.norm(avg_color[:3] - [0, 0, 0])
        dist_white = 1 - np.linalg.norm(avg_color[:3] - [255, 255, 255])

        background_color = (0, 0, 0) if dist_white > dist_black else (255, 255, 255)
        background = Image.new("RGB", svg_image.size, background_color)
        background.paste(svg_image, mask=svg_image.split()[3])
        opencv_template = cv2.cvtColor(np.array(background), cv2.COLOR_RGB2GRAY)
    else:
        opencv_template = cv2.cvtColor(np.array(svg_image), cv2.COLOR_RGB2GRAY)

    template_image = opencv_template
    source_image = opencv_image

    # Open the images
    # cv2.imshow('Image', opencv_image)
    # cv2.imshow('Template', opencv_template)
    # cv2.waitKey(0)
    # cv2.destroyAllWindows()

    # Initialize SIFT detector
    sift = cv2.SIFT_create()

    # Find the keypoints and descriptors with SIFT
    keypoints_1, descriptors_1 = sift.detectAndCompute(template_image, None)
    keypoints_2, descriptors_2 = sift.detectAndCompute(source_image, None)

    # Check if features are detected
    if descriptors_1 is None or descriptors_2 is None:
        raise ValueError(
            "Could not find enough features in one of the images. Try adjusting the SIFT parameters or using a different image."
        )

    # FLANN parameters and matcher
    FLANN_INDEX_KDTREE = 1
    index_params = dict(algorithm=FLANN_INDEX_KDTREE, trees=5)
    search_params = dict(checks=50)
    flann = cv2.FlannBasedMatcher(index_params, search_params)

    matches = flann.knnMatch(descriptors_1, descriptors_2, k=2)

    # Keep good matches: Lowe's ratio test
    good_matches = []
    for m, n in matches:
        if m and n and m.distance < 1 * n.distance:
            good_matches.append(m)

    # Draw all matches (for diagnostic purposes)
    img_matches = cv2.drawMatches(
        template_image,
        keypoints_1,
        source_image,
        keypoints_2,
        good_matches,
        None,
        flags=cv2.DrawMatchesFlags_NOT_DRAW_SINGLE_POINTS,
    )

    # Show all matches
    cv2.imshow("All Matches", img_matches)

    # Proceed only if there are enough good matches
    if len(good_matches) > 10:
        src_pts = np.float32(
            [keypoints_1[m.queryIdx].pt for m in good_matches]
        ).reshape(-1, 1, 2)
        dst_pts = np.float32(
            [keypoints_2[m.trainIdx].pt for m in good_matches]
        ).reshape(-1, 1, 2)

        # Compute Homography
        M, mask = cv2.findHomography(src_pts, dst_pts, cv2.RANSAC, 5.0)
        matchesMask = mask.ravel().tolist()

        # Get the dimensions of the template image
        h, w = template_image.shape[:2]
        pts = np.float32([[0, 0], [0, h - 1], [w - 1, h - 1], [w - 1, 0]]).reshape(
            -1, 1, 2
        )

        # Apply the homography to the template image's corners
        dst = cv2.perspectiveTransform(pts, M)

        # Draw bounding box in source image
        source_image = cv2.polylines(
            source_image, [np.int32(dst)], True, (0, 255, 0), 3, cv2.LINE_AA
        )
        cv2.imshow("Detected", source_image)
    else:
        print(f"Not enough good matches are found - {len(good_matches)}/10")
        matchesMask = None

    # Draw matches for visualization
    draw_params = dict(
        matchColor=(0, 255, 0), singlePointColor=None, matchesMask=matchesMask, flags=2
    )

    img_matches = cv2.drawMatches(
        template_image,
        keypoints_1,
        source_image,
        keypoints_2,
        good_matches,
        None,
        **draw_params,
    )

    # Show the matches
    cv2.imshow("Matches", img_matches)

    cv2.waitKey(0)
    cv2.destroyAllWindows()

    # Open all images used, using the systems image viewer (not plt)
    # print(pil_image.filename)
    # print(svg_image.filename)
    # os.system('open ' + pil_image.filename)
    # os.system('open ' + svg_image.filename)
    # try:
    #     os.startfile(pil_image.filename)
    #     os.startfile(svg_image.filename)
    # except AttributeError:
    #     import subprocess
    #     subprocess.call(['open', pil_image.filename])
    #     subprocess.call(['open', svg_image.filename])


from pytesseract import Output, pytesseract


def find_text_in_image(img, text):
    # Convert PIL Image to NumPy array
    img_array = np.array(img)

    # Convert the image to grayscale
    gray = cv2.cvtColor(img_array, cv2.COLOR_BGR2GRAY)

    # Use pytesseract to get the data from the image
    d = pytesseract.image_to_data(gray, output_type=Output.DICT)

    # Initialize an empty list to store the centers of the bounding boxes
    centers = []

    # Get the number of detected boxes
    n_boxes = len(d["level"])

    # Create a copy of the grayscale image to draw on
    img_draw = np.array(gray.copy())

    # Convert the img_draw grayscale image to RGB
    img_draw = cv2.cvtColor(img_draw, cv2.COLOR_GRAY2RGB)

    id = 0

    # Loop through each box
    for i in range(n_boxes):
        # Print the text of the box
        # If the text in the box matches the given text
        if text.lower() in d["text"][i].lower():
            # Find the start index of the matching text in the box
            start_index = d["text"][i].lower().find(text.lower())
            # Calculate the percentage of the box width that the start of the matching text represents
            start_percentage = start_index / len(d["text"][i])
            # Move the left edge of the box to the right by this percentage of the box width
            d["left"][i] = d["left"][i] + int(d["width"][i] * start_percentage)

            # Calculate the width of the matching text relative to the entire text in the box
            text_width_percentage = len(text) / len(d["text"][i])
            # Adjust the width of the box to match the width of the matching text
            d["width"][i] = int(d["width"][i] * text_width_percentage)

            # Calculate the center of the bounding box
            center = (
                d["left"][i] + d["width"][i] / 2,
                d["top"][i] + d["height"][i] / 2,
            )

            """bounding

            bounding

            bounding

            bounding

            bounding    bounding   bounding"""

            # Half both coordinates
            center = (center[0] / 2, center[1] / 2)

            # Add the center to the list
            centers.append(center)

            # Draw the bounding box on the image in red and make it slightly larger
            larger = 10
            cv2.rectangle(
                img_draw,
                (d["left"][i] - larger, d["top"][i] - larger),
                (
                    d["left"][i] + d["width"][i] + larger,
                    d["top"][i] + d["height"][i] + larger,
                ),
                (255, 0, 0),
                7,
            )

            # Create a small black square background for the ID
            cv2.rectangle(
                img_draw,
                (
                    d["left"][i] + d["width"][i] // 2 - larger * 2,
                    d["top"][i] + d["height"][i] // 2 - larger * 2,
                ),
                (
                    d["left"][i] + d["width"][i] // 2 + larger * 2,
                    d["top"][i] + d["height"][i] // 2 + larger * 2,
                ),
                (0, 0, 0),
                -1,
            )

            # Put the ID in the center of the bounding box in red
            cv2.putText(
                img_draw,
                str(id),
                (
                    d["left"][i] + d["width"][i] // 2 - larger,
                    d["top"][i] + d["height"][i] // 2 + larger,
                ),
                cv2.FONT_HERSHEY_DUPLEX,
                1,
                (255, 155, 155),
                4,
            )

            # Increment id
            id += 1

    if not centers:
        word_centers = []
        for word in text.split():
            for i in range(n_boxes):
                if word.lower() in d["text"][i].lower():
                    center = (
                        d["left"][i] + d["width"][i] / 2,
                        d["top"][i] + d["height"][i] / 2,
                    )
                    center = (center[0] / 2, center[1] / 2)
                    word_centers.append(center)

        for center1 in word_centers:
            for center2 in word_centers:
                if (
                    center1 != center2
                    and (
                        (center1[0] - center2[0]) ** 2 + (center1[1] - center2[1]) ** 2
                    )
                    ** 0.5
                    <= 400
                ):
                    centers.append(
                        ((center1[0] + center2[0]) / 2, (center1[1] + center2[1]) / 2)
                    )
                    break
            if centers:
                break

    bounding_box_image = Image.fromarray(img_draw)
    bounding_box_image.format = img.format

    # Debug by showing bounding boxes:
    # bounding_box_image.show()

    return centers, bounding_box_image



import time

import matplotlib.pyplot as plt
import numpy as np
import pyautogui
from PIL import Image

from ..utils.computer_vision import find_text_in_image


class Mouse:
    def __init__(self, computer):
        self.computer = computer

    def move(self, *args, x=None, y=None, index=None, svg=None):
        if len(args) > 1:
            raise ValueError(
                "Too many positional arguments provided: click(*args, x=None, y=None, show=True, index=None)"
            )
        elif len(args) == 1:
            text = args[0]
            # Take a screenshot
            img = self.computer.screenshot(show=False)

            # Find the text in the screenshot
            centers, bounding_box_image = find_text_in_image(img, text)

            # If the text was found
            if centers:
                # This could be refactored to be more readable
                if len(centers) > 1:
                    if index == None:
                        print(
                            f"(Message for language model) This text ('{text}') was found multiple times on screen. Please try 'click()' again, but pass in an `index` int to identify which one you want to click. The indices have been drawn on the image."
                        )
                        # Show the image using matplotlib
                        plt.imshow(np.array(bounding_box_image))
                        plt.show()
                        return
                    else:
                        center = centers[index]
                else:
                    center = centers[0]

                # Slowly move the mouse from its current position
                pyautogui.moveTo(center[0], center[1], duration=0.5)

            else:
                plt.imshow(np.array(bounding_box_image))
                plt.show()
                print("Your text was not found on the screen. Please try again.")
        elif x is not None and y is not None:
            # Move to the specified coordinates and click
            pyautogui.moveTo(x, y, duration=0.5)
        elif svg is not None:
            raise NotImplementedError("SVG handling not implemented yet.")
            # img = self.computer.screenshot(show=False)
            # # Move to the specified coordinates and click
            # coordinates = find_svg_in_image(svg, img)
            # if coordinates == None:
            #     print("Not found.")
            #     return
            # pyautogui.moveTo(coordinates[0], coordinates[1], duration=0.5)
            # pyautogui.click(coordinates[0], coordinates[1])
        else:
            raise ValueError("Either text or both x and y must be provided")

    def click(self, *args, **kwargs):
        if args or kwargs:
            self.move(*args, **kwargs)
        pyautogui.click()

    def down(self):
        pyautogui.mouseDown()

    def up(self):
        pyautogui.mouseUp()



import pyautogui
import pyperclip


class Clipboard:
    def get_selected_text(self):
        # Store the current clipboard content
        current_clipboard = pyperclip.paste()
        # Copy the selected text to clipboard
        pyautogui.hotkey("ctrl", "c", interval=0.15)
        # Get the selected text from clipboard
        selected_text = pyperclip.paste()
        # Reset the clipboard to its original content
        pyperclip.copy(current_clipboard)
        return selected_text



# import tiktoken
# from litellm import cost_per_token


# def count_tokens(text="", model="gpt-4"):
#     """
#     Count the number of tokens in a string
#     """

#     # Fix bug where models starting with openai/ for example can't find tokenizer
#     if '/' in model:
#         model = model.split('/')[-1]

#     # At least give an estimate if we can't find the tokenizer
#     try:
#         encoder = tiktoken.encoding_for_model(model)
#     except KeyError:
#         print(f"Could not find tokenizer for {model}. Defaulting to gpt-4 tokenizer.")
#         encoder = tiktoken.encoding_for_model("gpt-4")

#     return len(encoder.encode(text))


# def token_cost(tokens=0, model="gpt-4"):
#     """
#     Calculate the cost of the current number of tokens
#     """

#     (prompt_cost, _) = cost_per_token(model=model, prompt_tokens=tokens)

#     return round(prompt_cost, 6)


# def count_messages_tokens(messages=[], model=None):
#     """
#     Count the number of tokens in a list of messages
#     """

#     tokens_used = 0

#     for message in messages:
#         if isinstance(message, str):
#             tokens_used += count_tokens(message, model=model)
#         elif "message" in message:
#             tokens_used += count_tokens(message["message"], model=model)

#             if "code" in message:
#                 tokens_used += count_tokens(message["code"], model=model)

#             if "output" in message:
#                 tokens_used += count_tokens(message["output"], model=model)

#     prompt_cost = token_cost(tokens_used, model=model)

#     return (tokens_used, prompt_cost)



from .languages.applescript import AppleScript
from .languages.html import HTML
from .languages.javascript import JavaScript
from .languages.powershell import PowerShell
from .languages.python import Python
from .languages.r import R
from .languages.react import React
from .languages.shell import Shell


class Terminal:
    def __init__(self):
        self.languages = [
            Python,
            Shell,
            JavaScript,
            HTML,
            AppleScript,
            R,
            PowerShell,
            React,
        ]
        self._active_languages = {}

    def get_language(self, language):
        for lang in self.languages:
            if language.lower() == lang.name.lower() or (
                hasattr(lang, "aliases") and language in lang.aliases
            ):
                return lang
        return None

    def run(self, language, code):
        if language not in self._active_languages:
            self._active_languages[language] = self.get_language(language)()
        try:
            yield from self._active_languages[language].run(code)
        except GeneratorExit:
            self.stop()

    def stop(self):
        for language in self._active_languages.values():
            language.stop()

    def terminate(self):
        for language_name in list(self._active_languages.keys()):
            language = self._active_languages[language_name]
            if (
                language
            ):  # Not sure why this is None sometimes. We should look into this
                language.terminate()
            del self._active_languages[language_name]



import os
import platform
import random
import time

import pyautogui


class Keyboard:
    def write(self, text):
        pyautogui.write(text)

    def press(self, keys):
        pyautogui.press(keys)

    def hotkey(self, *args):
        modifiers = {"command", "control", "option", "shift"}
        if "darwin" in platform.system().lower() and len(args) == 2:
            # pyautogui.hotkey seems to not work, so we use applescript
            # Determine which argument is the keystroke and which is the modifier
            keystroke, modifier = args if args[0] not in modifiers else args[::-1]

            # Create the AppleScript
            script = f"""
            tell application "System Events"
                keystroke "{keystroke}" using {modifier}
            end tell
            """

            # Execute the AppleScript
            os.system("osascript -e '{}'".format(script))
        else:
            pyautogui.hotkey(*args, interval=0.15)

    def down(self, key):
        pyautogui.keyDown(key)

    def up(self, key):
        pyautogui.keyUp(key)
